from django.views.i18n import javascript_catalog
from django.views.static import serve
from django.conf import settings
from django.conf.urls import include, url
from django.contrib import admin
from djangofloor.views import robots
from djangofloor.scripts import load_celery

from autoplanner import views


__author__ = 'Matthieu Gallet'
load_celery()
admin.autodiscover()
admin_urls = admin.site.urls

urlpatterns = [url(r'^jsi18n/$', javascript_catalog, {'packages': ('djangofloor', 'django.contrib.admin', ), }),
               url(r'^' + settings.MEDIA_URL[1:] + '(?P<path>.*)$', serve, {'document_root': settings.MEDIA_ROOT}),
               url(r'^' + settings.STATIC_URL[1:] + '(?P<path>.*)$', serve, {'document_root': settings.STATIC_ROOT}),
               url(r'^robots\.txt$', robots),
               url('^org/(?P<organization_pk>\d+)\.html', views.organization, name='organization'),
               url('^org/start/(?P<organization_pk>\d+)\.html', views.schedule_task, name='schedule_start'),
               url('^org/cancel/(?P<organization_pk>\d+)\.html', views.cancel_schedule_task, name='cancel_schedule'),
               url('^org/ical/(?P<organization_pk>\d+)/all/(?P<title>.*)\.ics', views.generate_ics, name='ical'),
               url('^org/ical/(?P<organization_pk>\d+)/agent/(?P<agent_pk>\d+)/(?P<title>.*)\.ics',
                   views.generate_ics, name='ical'),
               url('^org/ical/(?P<organization_pk>\d+)/cat/(?P<category_pk>\d+)/(?P<title>.*)\.ics',
                   views.generate_ics, name='ical'),
               url('^tasks/multiply/(?P<task_pk>\d+)\.html', views.multiply_task, name='multiply_task'),
               url('^tasks/apply_schedule_run/(?P<schedule_run_pk>\d+)\.html', views.apply_schedule_run,
                   name='apply_schedule_run'),
               url(r'^', include(admin_urls[:2])),
               url(r'^organization/(?P<organization_pk>\d+)/$', views.organization_index, name='organization_index'),
               url(r'^$', views.index, name='index'),
               ]

if settings.DEBUG and settings.USE_DEBUG_TOOLBAR:
    import debug_toolbar
    urlpatterns += [url(r'^__debug__/', include(debug_toolbar.urls)), ]



import datetime
import random

from django.conf import settings
from django.core.validators import RegexValidator
from django.db import models
from django.db.models import Q
from django.http import HttpRequest
from django.urls import reverse
from django.utils.formats import date_format, time_format
from django.utils.text import force_text
from django.utils.timezone import get_default_timezone
from django.utils.translation import ugettext_lazy as _

API_KEY_VARIABLE = 'api_key'

__author__ = 'Matthieu Gallet'


def default_day_start():
    x = datetime.datetime.now(tz=get_default_timezone())
    return datetime.datetime(year=x.year, month=x.month, day=x.day, hour=0, minute=0, second=0,
                             tzinfo=get_default_timezone())


def default_day_end():
    return default_day_start() + datetime.timedelta(hours=23, minutes=59, seconds=59)


def default_token(length=64, allowed_chars='abcdefghijklmnopqrstuvwxyz'
                                           'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'):
    # noinspection PyUnusedLocal
    return ''.join([random.choice(allowed_chars) for i in range(length)])


class Organization(models.Model):
    name = models.CharField(_('Name'), db_index=True, max_length=500)
    message = models.CharField(_('Message'), blank=True, max_length=500, default='')
    description = models.TextField(_('Description'), default='', blank=True)
    celery_task_id = models.CharField(_('Celery task id'), db_index=True, max_length=60, blank=True, null=True,
                                      default=None)
    celery_start = models.DateTimeField(_('Celery start'), null=True, blank=True, default=None)
    celery_end = models.DateTimeField(_('Celery end'), null=True, blank=True, default=None)
    access_token = models.CharField(_('Access token'), default=default_token, max_length=300,
                                    validators=[RegexValidator(r'^[a-zA-Z0-9]{1,300}$')])
    max_compute_time = models.PositiveIntegerField(_('Maximum time, in seconds, for finding a solution'),
                                                   default=None, blank=True, null=True,
                                                   help_text=_('Leave it blank if you do not want to set a limit'))
    admins = models.ManyToManyField(settings.AUTH_USER_MODEL, db_index=True, verbose_name=_('Administrators'))
    current_schedule = models.ForeignKey('ScheduleRun', default=None, blank=True, null=True, db_index=True,
                                         related_name='current_organizations', on_delete=models.SET_NULL)

    @classmethod
    def query(cls, request: HttpRequest, readonly=False):
        access_token = ''
        user = getattr(request, 'user')
        if hasattr(request, 'GET'):
            # noinspection PyCallByClass,PyTypeChecker
            access_token = request.GET.get(API_KEY_VARIABLE, '')
        if not user or user.is_anonymous:
            return cls.objects.filter(access_token=access_token)
        # noinspection PyUnresolvedReferences
        elif user.is_superuser:
            return cls.objects.all()
        elif readonly:
            # noinspection PyUnresolvedReferences
            return cls.objects.filter(Q(admins=user) | Q(access_token=access_token))
        # noinspection PyUnresolvedReferences
        return cls.objects.filter(admins=user)

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return '%s?%s=%s' % (reverse('organization', kwargs={'organization_pk': self.pk}),
                             API_KEY_VARIABLE, self.access_token)


class ScheduleRun(models.Model):
    organization = models.ForeignKey(Organization, db_index=True, on_delete=models.CASCADE)
    status = models.NullBooleanField(_('Is valid?'), db_index=True, default=None)
    is_selected = models.BooleanField(_('Selected?'), db_index=True, default=False)
    message = models.TextField(_('Result'), max_length=500, default='', blank=True)
    celery_task_id = models.CharField(_('Celery task id'), db_index=True, max_length=60, blank=True, null=True,
                                      default=None)
    celery_start = models.DateTimeField(_('Computation start'), null=True, blank=True, default=None)
    celery_end = models.DateTimeField(_('Computation end'), null=True, blank=True, default=None)
    process_id = models.IntegerField(_('Process ID'), db_index=True, blank=True, null=True, default=None)
    result_dict = models.TextField(_('JSON-serialized result'), blank=True, default=None, null=True)

    def __str__(self):
        end = self.celery_end
        if not end:
            return force_text(_('Not finished yet.'))
        return '%(d)s, %(t)s' % {'d': date_format(end, use_l10n=True), 't': time_format(end, use_l10n=True)}

    def __unicode__(self):
        return self.__str__()


class OrganizationObject(models.Model):
    organization = models.ForeignKey(Organization, db_index=True, on_delete=models.CASCADE)

    class Meta(object):
        abstract = True

    @classmethod
    def query(cls, request: HttpRequest):
        # noinspection PyUnresolvedReferences
        if request.user.is_superuser:
            return cls.objects.all()
        # noinspection PyUnresolvedReferences
        return cls.objects.filter(organization__admins=request.user)


class Agent(OrganizationObject):
    name = models.CharField(_('Name'), db_index=True, max_length=500)
    start_time = models.DateTimeField(_('Arrival time'), db_index=True, default=None, blank=True, null=True,
                                      help_text=_('Before this date, the agent cannot perform'
                                                  'any task.'))
    end_time = models.DateTimeField(_('Leaving time'), db_index=True, default=None, blank=True, null=True,
                                    help_text=_('After this date, the agent cannot perform any task.'))

    class Meta(object):
        ordering = ('name',)

    def __str__(self):
        return self.name


class Category(OrganizationObject):
    BALANCE_TIME = 'time'
    BALANCE_NUMBER = 'number'
    name = models.CharField(_('Name'), db_index=True, max_length=500)
    balancing_mode = models.CharField(_('Balancing mode'),
                                      max_length=10, choices=((None, _('No balancing')),
                                                              (BALANCE_TIME, _('Total task time')),
                                                              (BALANCE_NUMBER, _('Total task number'))),
                                      blank=True, null=True, default=None)
    balancing_tolerance = models.FloatField(_('Tolerance for balancing the total duration (in seconds)|tasks '
                                              'across agents'), default=0., blank=True, null=True)
    auto_affinity = models.FloatField(_('Affinity for allocating successive tasks of the same category '
                                        'to the same agent'), default=0., blank=True)

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _('Category of tasks')
        verbose_name_plural = _('Categories of tasks')

    @property
    def balancing_tolerance_as_timedelta_2(self):
        return datetime.timedelta(seconds=(self.balancing_tolerance or 0))

    @property
    def balancing_tolerance_2(self):
        return self.balancing_tolerance or 0


class MaxAffectation(OrganizationObject):
    MINIMUM = 'min'
    MAXIMUM = 'max'
    category = models.ForeignKey(Category, db_index=True, on_delete=models.CASCADE)
    mode = models.CharField(_('Mode'), max_length=3, choices=((MINIMUM, _('At least this number of tasks')),
                                                              (MAXIMUM, _('At most this number of tasks'))),
                            default=MAXIMUM)
    range_time_slice_days = models.IntegerField(_('Period length (days)'), default=2)
    range_time_slice_hours = models.IntegerField(_('Period length (hours)'), default=0)
    range_time_slice_seconds = models.IntegerField(_('Period length (seconds)'), default=0)

    @property
    def range_time_slice(self):
        return datetime.timedelta(days=self.range_time_slice_days, hours=self.range_time_slice_hours,
                                  seconds=self.range_time_slice_seconds)

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        if self.organization_id is None and self.category_id:
            self.organization_id = self.category.organization_id
        super().save(force_insert=force_insert, force_update=force_update, using=using,
                     update_fields=update_fields)

    class Meta:
        abstract = True


class MaxTaskAffectation(MaxAffectation):
    task_maximum_count = models.IntegerField(_('Number of tasks in this range'), default=1)

    class Meta:
        verbose_name = _('Number of tasks performed by an agent in a category')
        verbose_name_plural = _('Number of tasks performed by an agent in a category')


class MaxTimeTaskAffectation(MaxAffectation):
    task_maximum_time_days = models.IntegerField(_('Total task time in this range (days)'), default=1)
    task_maximum_time_hours = models.IntegerField(_('Total task time in this range (hours)'), default=0)
    task_maximum_time_seconds = models.IntegerField(_('Total task time in this range (seconds)'), default=0)

    @property
    def task_maximum_time(self):
        return datetime.timedelta(days=self.task_maximum_time_days, hours=self.task_maximum_time_hours,
                                  seconds=self.task_maximum_time_seconds)

    class Meta:
        verbose_name = _('Maximum time spent by an agent in a category')
        verbose_name_plural = _('Maximum time spent by an agent in a category')


class Task(OrganizationObject):
    orders = {'name', '-name', 'start_time', '-start_time', 'end_time', '-end_time'}
    categories = models.ManyToManyField(Category, db_index=True, blank=True)
    name = models.CharField(_('Name'), db_index=True, max_length=500)
    start_time = models.DateTimeField(_('Start time'), db_index=True, default=default_day_start)
    end_time = models.DateTimeField(_('End time'), db_index=True, default=default_day_end)
    agent = models.ForeignKey(Agent, db_index=True, null=True, default=None, blank=True, on_delete=models.SET_NULL)
    fixed = models.BooleanField(_('Forced agent'), db_index=True, default=False)
    task_serie = models.ForeignKey('self', db_index=True, null=True, default=None, on_delete=models.SET_NULL,
                                   blank=True)

    class Meta(object):
        ordering = ('start_time', 'end_time',)

    @property
    def duration(self) -> datetime.timedelta:
        return self.end_time - self.start_time

    def __str__(self):
        start = date_format(self.start_time, use_l10n=True)
        end = date_format(self.end_time, use_l10n=True)
        return '%s (%s -> %s)' % (self.name, start, end)


class AgentCategoryPreferences(OrganizationObject):
    category = models.ForeignKey(Category, db_index=True, on_delete=models.CASCADE)
    agent = models.ForeignKey(Agent, db_index=True, on_delete=models.CASCADE)
    affinity = models.FloatField(_('Affinity of the agent for the category.'), default=0., blank=True)
    balancing_offset = models.FloatField(_('Number of time units already done'), default=0, blank=True)
    balancing_count = models.FloatField(_('If a task of this category performed by this agent counts twice, '
                                          'set this number to 2.0.'),
                                        default=1.0, blank=True, null=True,
                                        help_text=_('Blank if the agent cannot perform tasks of this category'))

    class Meta(object):
        verbose_name = 'preference by agent and category'
        verbose_name_plural = 'preferences by agent and category'

    @property
    def offset_as_timedelta(self):
        return datetime.timedelta(seconds=self.balancing_offset)

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        if self.organization_id is None and self.category_id:
            self.organization_id = self.category.organization_id
        elif self.organization_id is None and self.agent_id:
            self.organization_id = self.agent.organization_id
        super().save(force_insert=force_insert, force_update=force_update, using=using,
                     update_fields=update_fields)


class AgentTaskExclusion(OrganizationObject):
    agent = models.ForeignKey(Agent, db_index=True, on_delete=models.CASCADE)
    task = models.ForeignKey(Task, db_index=True, verbose_name=_('Task'), on_delete=models.CASCADE,
                             help_text=_('Select the task that cannot be performed by the agent.'))

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        if self.organization_id is None and self.task_id:
            self.organization_id = self.task.organization_id
        elif self.organization_id is None and self.agent_id:
            self.organization_id = self.agent.organization_id
        super().save(force_insert=force_insert, force_update=force_update, using=using,
                     update_fields=update_fields)



import datetime
from django.test import TestCase
from django.utils.timezone import utc
from autoplanner.models import Organization, Agent, Category, Task
from autoplanner.schedule import Scheduler

__author__ = 'Matthieu Gallet'


class BaseTest(TestCase):
    def get_organization(self):
        org = Organization(name='O')
        org.save()
        agent_1 = Agent(organization=org, name='A1')
        agent_1.save()
        agent_2 = Agent(organization=org, name='A2')
        agent_2.save()
        agent_3 = Agent(organization=org, name='A3')
        agent_3.save()
        category_0 = Category(organization=org, name='C0')
        category_0.save()
        category_1 = Category(organization=org, name='C1', parent_category=category_0)
        category_1.save()
        category_2 = Category(organization=org, name='C2', parent_category=category_0)
        category_2.save()
        return org


class TestSimple(BaseTest):
    def test_empty(self):
        org = self.get_organization()
        s = Scheduler(org)
        self.assertEqual(['min:'], [str(x) for x in s.constraints()])

    def get_time(self, value):
        return datetime.datetime(2016, 1, 1, value, 0, 0, tzinfo=utc)

    def test_single(self):
        org = self.get_organization()
        category_1 = org.category_set.get(name='C1')
        task_1 = Task(organization=org, name='E1', start_time=self.get_time(0), end_time=self.get_time(1),
                      category=category_1)
        task_1.save()
        # s = Scheduler(org)
        # for c in s.constraints():
        # print(c)

        org = self.get_organization()
        category_1 = org.category_set.get(name='C1')
        category_1.balancing_mode = Category.BALANCE_NUMBER
        category_1.save()
        task_1 = Task(organization=org, name='E1', start_time=self.get_time(0), end_time=self.get_time(1),
                      category=category_1)
        task_1.save()
        s = Scheduler(org)
        result_list = s.solve(verbose=False)
        result_dict = s.result_by_agent(result_list)
        self.assertEqual({4: {2}}, result_dict)

    def test_simple_balance(self):
        org = self.get_organization()
        category_1 = org.category_set.get(name='C1')
        assert isinstance(category_1, Category)
        category_1.balancing_mode = Category.BALANCE_NUMBER
        category_1.balancing_tolerance = 0
        category_1.save()
        task_1 = Task(organization=org, name='E1', start_time=self.get_time(0), end_time=self.get_time(1),
                      category=category_1)
        task_1.save()
        task_2 = Task(organization=org, name='E2', start_time=self.get_time(1), end_time=self.get_time(2),
                      category=category_1)
        task_2.save()
        task_3 = Task(organization=org, name='E3', start_time=self.get_time(2), end_time=self.get_time(3),
                      category=category_1)
        task_3.save()
        s = Scheduler(org)
        result_list = s.solve(verbose=False)
        result_dict = s.result_by_agent(result_list)
        self.assertEqual({1: {3}, 2: {2}, 3: {1}}, result_dict)

    def test_infeasible(self):
        org = self.get_organization()
        category_1 = org.category_set.get(name='C1')
        assert isinstance(category_1, Category)
        category_1.balancing_mode = Category.BALANCE_NUMBER
        category_1.balancing_tolerance = 0
        category_1.save()
        task_1 = Task(organization=org, name='E1', start_time=self.get_time(0), end_time=self.get_time(1),
                      category=category_1)
        task_1.save()
        task_2 = Task(organization=org, name='E2', start_time=self.get_time(1), end_time=self.get_time(2),
                      category=category_1)
        task_2.save()
        task_3 = Task(organization=org, name='E3', start_time=self.get_time(2), end_time=self.get_time(3),
                      category=category_1)
        task_3.save()
        task_4 = Task(organization=org, name='E4', start_time=self.get_time(3), end_time=self.get_time(4),
                      category=category_1)
        task_4.save()
        s = Scheduler(org)
        result_list = s.solve(verbose=False)
        result_dict = s.result_by_agent(result_list)
        self.assertEqual({}, result_dict)
        category_1.balancing_tolerance = 1
        category_1.save()
        s = Scheduler(org)
        result_list = s.solve(verbose=False)
        result_dict = s.result_by_agent(result_list)
        self.assertEqual({1: {1, 2}, 2: {3}, 3: {4}}, result_dict)

    def test_overlap(self):
        org = self.get_organization()
        category_1 = org.category_set.get(name='C1')
        Agent.objects.filter(organization=org, name__in=['A2', 'A3']).delete()
        task_1 = Task(organization=org, name='E1', start_time=self.get_time(10), end_time=self.get_time(20),
                      category=category_1)
        task_1.save()
        task_2 = Task(organization=org, name='E2', start_time=self.get_time(15), end_time=self.get_time(23),
                      category=category_1)
        task_2.save()
        s = Scheduler(org)
        result_list = s.solve(verbose=False)
        result_dict = s.result_by_agent(result_list)
        self.assertEqual({}, result_dict)

        task_2.start_time = self.get_time(21)
        task_2.save()
        s = Scheduler(org)
        result_list = s.solve(verbose=False)
        result_dict = s.result_by_agent(result_list)
        self.assertEqual({1: {1, 2}}, result_dict)

        task_2.start_time = self.get_time(20)
        task_2.save()
        s = Scheduler(org)
        result_list = s.solve(verbose=False)
        result_dict = s.result_by_agent(result_list)
        self.assertEqual({1: {1, 2}}, result_dict)

        task_2.start_time = self.get_time(5)
        task_2.end_time = self.get_time(10)
        task_2.save()
        s = Scheduler(org)
        result_list = s.solve(verbose=False)
        result_dict = s.result_by_agent(result_list)
        self.assertEqual({1: {1, 2}}, result_dict)

        task_2.start_time = self.get_time(5)
        task_2.end_time = self.get_time(11)
        task_2.save()
        s = Scheduler(org)
        result_list = s.solve(verbose=False)
        result_dict = s.result_by_agent(result_list)
        self.assertEqual({}, result_dict)

        task_2.start_time = self.get_time(5)
        task_2.end_time = self.get_time(23)
        task_2.save()
        s = Scheduler(org)
        result_list = s.solve(verbose=False)
        result_dict = s.result_by_agent(result_list)
        self.assertEqual({}, result_dict)

        task_2.start_time = self.get_time(12)
        task_2.end_time = self.get_time(17)
        task_2.save()
        s = Scheduler(org)
        result_list = s.solve(verbose=False)
        result_dict = s.result_by_agent(result_list)
        self.assertEqual({}, result_dict)



import re
import subprocess
import tempfile

import itertools
from django.conf import settings

from autoplanner.models import Organization, MaxTimeTaskAffectation, Task, ScheduleRun

__author__ = 'Matthieu Gallet'


class Constraint(object):
    def __init__(self, value: str, name: str=''):
        self.value = value
        self.name = name

    def __str__(self):
        return self.value


class Scheduler(object):
    def __init__(self, organization: Organization):
        self.organization = organization
        self.agents = {x for x in organization.agent_set.all()}
        self.categories = {x for x in organization.category_set.all()}
        self.categories_by_pk = {x.pk: x for x in self.categories}
        self.tasks = {x for x in organization.task_set.all() if (x.start_time < x.end_time)}
        self.agent_category_preferences = {x for x in organization.agentcategorypreferences_set.all()}

        self.task_durations = {task.pk: (task.end_time - task.start_time).total_seconds()
                               for task in self.tasks}
        self.agent_pks = {x.pk for x in self.agents}
        # self.agent_pks = {agent1.pk, agent2.pk, agent3.pk}
        self.categories_by_task = self.get_categories_by_task()
        # self.categories_by_task[task.pk] = {category1.pk, category2.pk, category3.pk}
        self.tasks_by_category = self.get_tasks_by_category()
        # self.tasks_by_categories[category.pk] = {task1, task2, task3}
        self.agent_exclusions_by_category = self.get_agent_exclusions_by_category()
        # self.agent_exclusions_by_category[category.pk] = {agent1.pk, agent2.pk, agent3.pk}
        self.agent_exclusions_by_task = self.get_agent_exclusions_by_task()
        # self.agent_exclusions_by_task[task.pk] = {agent1.pk, agent2.pk, agent3.pk}
        self.preferences_by_agent_by_category = self.get_preferences_by_agent_by_category()
        # self.preferences_by_agent_by_category[category.pk][agent.pk] = (balancing_offset, balancing_count, affinity)
        self.max_task_affectations_by_category = self.get_max_task_affectations_by_category()
        # self.max_task_affectations_by_category[category.pk] = [max_task_affectation_1, max_task_affectation_2]
        self.available_agents_by_tasks = {task.pk: (self.agent_pks - self.agent_exclusions_by_task[task.pk])
                                          for task in self.tasks}

    def get_categories_by_task(self):
        result = {x.pk: set() for x in self.tasks}
        for related in Task.categories.through.objects.filter(category__id__in=self.categories_by_pk):
            result[related.task_id].add(related.category_id)
        return result

    def get_tasks_by_category(self):
        result = {x.pk: set() for x in self.categories}
        tasks_by_pk = {x.pk: x for x in self.tasks}
        for task_pk, category_pks in self.categories_by_task.items():
            for category_pk in category_pks:
                result[category_pk].add(tasks_by_pk[task_pk])
        return result

    def get_agent_exclusions_by_category(self):
        agent_exclusions_by_category = {category.pk: set() for category in self.categories}
        for a in self.agent_category_preferences:
            if a.balancing_count is None:
                agent_exclusions_by_category[a.category_id].add(a.agent_id)
        return agent_exclusions_by_category

    def get_agent_exclusions_by_task(self):
        agent_task_exclusions = {task.pk: set() for task in self.tasks}
        for task in self.tasks:
            task_pk = task.pk
            for category_pk in self.categories_by_task[task_pk]:
                for agent_pk in self.agent_exclusions_by_category[category_pk]:
                    agent_task_exclusions[task_pk].add(agent_pk)
            for agent in self.agents:
                # TODO optimize
                if agent.start_time is not None and agent.start_time > task.start_time:
                    agent_task_exclusions[task_pk].add(agent.pk)
                elif agent.end_time is not None and agent.end_time < task.end_time:
                    agent_task_exclusions[task_pk].add(agent.pk)
        for agent_task_exclusion in self.organization.agenttaskexclusion_set.all():
            agent_task_exclusions[agent_task_exclusion.task_pk].add(agent_task_exclusion.agent_id)
        return agent_task_exclusions

    def get_preferences_by_agent_by_category(self):
        preferences = {c.pk: {} for c in self.categories}
        for a in self.agent_category_preferences:
            preferences[a.category_id][a.agent_id] = (a.balancing_offset, a.balancing_count, a.affinity)
        return preferences

    def get_max_task_affectations_by_category(self):
        max_task_affectations_by_category = {x.pk: [] for x in self.categories}
        for max_affectation in self.organization.maxtaskaffectation_set.all():
            max_task_affectations_by_category[max_affectation.category_id].append(max_affectation)
        for max_affectation in self.organization.maxtimetaskaffectation_set.all():
            max_task_affectations_by_category[max_affectation.category_id].append(max_affectation)
        return max_task_affectations_by_category

    def apply_max_task_affectations(self, category_pk: int):
        """ "Agent A cannot execute more than X tasks of category C in less than T time slices"
        :param category_pk:
        """
        task_data = [(task.pk, task.start_time, task.end_time) for task in self.tasks_by_category[category_pk]]
        task_data.sort(key=lambda x: (x[1], x[2]))
        agent_pks = self.agent_pks - self.agent_exclusions_by_category[category_pk]
        previous_start_time = None
        for index, data in enumerate(task_data):
            task_pk, start_time, __ = data
            if start_time == previous_start_time:
                continue
            current_tasks = {task_pk}
            for max_affectation in self.max_task_affectations_by_category[category_pk]:
                end_block_time_slice = start_time + max_affectation.range_time_slice
                for data_2 in task_data[index + 1:]:
                    task_pk_2, start_time_2, __ = data_2
                    if start_time_2 >= end_block_time_slice:
                        break
                    current_tasks.add(task_pk_2)
                direction = '<=' if max_affectation.mode == max_affectation.MAXIMUM else '>='
                if isinstance(max_affectation, MaxTimeTaskAffectation):
                    for agent_pk in agent_pks:
                        variables = ['%d * %s' % (self.task_durations[task_pk], self.variable(agent_pk, task_pk))
                                     for task_pk in current_tasks]
                        yield Constraint('%s %s %d' % (' + '.join(variables), direction,
                                                       max_affectation.task_maximum_time))
                else:
                    for agent_pk in agent_pks:
                        variables = [self.variable(agent_pk, task_pk) for task_pk in current_tasks]
                        yield Constraint('%s %s %d' % (' + '.join(variables), direction,
                                                       max_affectation.task_maximum_count))
            previous_start_time = start_time

    def apply_all_tasks_must_be_done(self):
        """ "Exactly one agent must perform each task"
        """
        for task_pk, task_agent_pks in self.available_agents_by_tasks.items():
            yield Constraint('%s = 1' % ' + '.join([self.variable(agent_pk, task_pk) for agent_pk in task_agent_pks]))
            if len(task_agent_pks) == len(self.agent_pks):
                continue
            yield Constraint('%s = 1' % ' + '.join([self.variable(agent_pk, task_pk) for agent_pk in self.agent_pks]))

    def apply_fixed_tasks(self):
        """ "Task E must be performed by agent A" """
        for task in self.tasks:
            if task.fixed and task.agent_id:
                yield Constraint('%s = 1' % self.variable(task.agent_id, task.pk))

    def apply_single_task_per_agent(self):
        """ "Agent X can perform at most one task (of the same category) at a time" """
        current_tasks = set()
        resumed_tasks = {}
        for task in self.tasks:
            resumed_tasks.setdefault(task.start_time, []).append((task.pk, current_tasks.add))
            resumed_tasks.setdefault(task.end_time, []).append((task.pk, current_tasks.remove))
        time_slices = [x for x in resumed_tasks]
        time_slices.sort()
        for time_slice in time_slices:
            for (task_pk, action) in resumed_tasks[time_slice]:
                action(task_pk)
            if not current_tasks:
                continue
            current_tasks_by_category = {}
            for task_pk in current_tasks:
                for category_pk in self.categories_by_task[task_pk]:
                    current_tasks_by_category.setdefault(category_pk, set()).add(task_pk)
            for agent_pk in self.agent_pks:
                for current_task_subset in current_tasks_by_category.values():
                    variables = [self.variable(agent_pk, task_pk) for task_pk in current_task_subset]
                    yield Constraint('%s <= 1' % ' + '.join(variables))

    def apply_balancing_constraints(self):
        for category in self.categories:
            if category.balancing_mode is None or category.balancing_tolerance is None:
                continue
            category_pk = category.pk
            cat_task_pks = [task.pk for task in self.tasks_by_category[category_pk]]
            cat_agent_pks = self.agent_pks - self.agent_exclusions_by_category[category_pk]
            for agent_pk in cat_agent_pks:
                agent_preferences = self.preferences_by_agent_by_category[category_pk].get(agent_pk, (0, 1., 0.))
                if category.balancing_mode == category.BALANCE_NUMBER:
                    ag_sum = ['%g * %s' % (agent_preferences[1], self.variable(agent_pk, task_pk))
                              for task_pk in cat_task_pks]
                else:
                    ag_sum = ['%g * %d * %s' % (agent_preferences[1], self.task_durations[task_pk],
                                                self.variable(agent_pk, task_pk)) for task_pk in cat_task_pks]
                yield Constraint('%g + %s = %s' % (agent_preferences[0] * agent_preferences[1], ' + '.join(ag_sum),
                                                   self.category_variable(category_pk, agent_pk)))
            # ag_sum = [self.category_variable(category_pk, agent_pk) for agent_pk in cat_agent_pks]
            # yield Constraint('%s = %s' % (' + '.join(ag_sum), self.category_variable(category_pk)))
            for agent_pk_1, agent_pk_2 in itertools.product(cat_agent_pks, cat_agent_pks):
                if agent_pk_1 >= agent_pk_2:
                    continue
                yield Constraint(
                    '%s <= %s + %g' % (self.category_variable(category_pk, agent_pk_1),
                                       self.category_variable(category_pk, agent_pk_2), category.balancing_tolerance))
                yield Constraint(
                    '%s <= %s + %g' % (self.category_variable(category_pk, agent_pk_2),
                                       self.category_variable(category_pk, agent_pk_1), category.balancing_tolerance))
            # for agent_pk in cat_agent_pks:
            #     count = len(cat_agent_pks)
            #     yield Constraint(
            #         '%d * %s <= %s + %g' % (count, self.category_variable(category_pk, agent_pk),
            #                               self.category_variable(category_pk), category.balancing_tolerance * count))
            #     yield Constraint(
            #         '%d * %s >= %s - %g' % (count, self.category_variable(category_pk, agent_pk),
            #                                 self.category_variable(category_pk), category.balancing_tolerance * count)
            #     )

    def compute_balancing(self, result_list):
        """Return a dict
        :return {category.pk: (name, mode, max_value_among_all_agents - min_value_among_all_agents)}"""
        balances = {}

        result_dict = {}
        for agent_pk, task_pk in result_list:
            result_dict.setdefault(agent_pk, set()).add(task_pk)

        def is_affected_to(agent_pk_, task_pk_):
            return 1 if (agent_pk_ in result_dict and task_pk_ in result_dict[agent_pk_]) else 0

        for category in self.categories:
            if category.balancing_mode is None or category.balancing_tolerance is None:
                continue
            category_pk = category.pk
            cat_task_pks = [task.pk for task in self.tasks_by_category[category_pk]]
            cat_agent_pks = self.agent_pks - self.agent_exclusions_by_category[category_pk]
            cat_min = None
            cat_max = None
            for agent_pk in cat_agent_pks:
                agent_preferences = self.preferences_by_agent_by_category[category_pk].get(agent_pk, (0, 1., 0.))
                if category.balancing_mode == category.BALANCE_NUMBER:
                    ag_sum = sum((agent_preferences[1] * is_affected_to(agent_pk, task_pk)) for task_pk in cat_task_pks)
                else:
                    ag_sum = sum((agent_preferences[1] * self.task_durations[task_pk] *
                                  is_affected_to(agent_pk, task_pk)) for task_pk in cat_task_pks)
                ag_offset = agent_preferences[0] * agent_preferences[1]
                ag_total = ag_sum + ag_offset
                if cat_min is None or cat_min > ag_total:
                    cat_min = ag_total
                if cat_max is None or cat_max < ag_total:
                    cat_max = ag_total
            if cat_min is not None:
                balances[category_pk] = (category.name, category.balancing_mode, cat_max - cat_min)
        return balances

    def apply_affinity_constraints(self):
        variables = []
        for category in self.categories:
            category_pk = category.pk
            cat_task_pks = [(task.pk, task.start_time, task.end_time) for task in self.tasks_by_category[category_pk]]
            # if category.auto_affinity:
            # cat_task_pks.sort(key=lambda x: x[1])
            for agent_pk in self.agent_pks - self.agent_exclusions_by_category[category_pk]:
                agent_preferences = self.preferences_by_agent_by_category[category_pk].get(agent_pk, (0, 1., 0.))
                variables += ['%g * %s' % (agent_preferences[2], self.variable(agent_pk, x[0]))
                              for x in cat_task_pks if agent_preferences[2]]
        if variables:
            yield Constraint('min: -%s' % self.affinity_variable())
            yield Constraint('%s = %s' % (' + '.join(variables), self.affinity_variable()))
        else:
            yield Constraint('min:')

    def constraints(self):
        yield from self.apply_affinity_constraints()
        # all tasks must be done
        yield from self.apply_all_tasks_must_be_done()
        # task with a fixed agent
        yield from self.apply_fixed_tasks()
        # at most one task by agent at the same time
        yield from self.apply_single_task_per_agent()
        for category in self.categories:
            if self.max_task_affectations_by_category[category.pk]:
                yield from self.apply_max_task_affectations(category.pk)
        yield from self.apply_balancing_constraints()
        for task_pk, agent_pks in self.available_agents_by_tasks.items():
            for agent_pk in agent_pks:
                yield Constraint('%s >= 0' % self.variable(agent_pk, task_pk))
                yield Constraint('%s <= 1' % self.variable(agent_pk, task_pk))
        for task_pk, agent_pks in self.available_agents_by_tasks.items():
            for agent_pk in agent_pks:
                yield Constraint('int %s' % self.variable(agent_pk, task_pk))

    @staticmethod
    def variable(agent_pk, task_pk):
        return 'v_a%s_e%s' % (agent_pk, task_pk)

    result_line_re = r'^v_a(\d+)_e(\d+)\s+1$'

    @staticmethod
    def category_variable(category_pk, agent_pk=None):
        if agent_pk is None:
            return 'c_c%s' % category_pk
        return 'c_c%s_a%s' % (category_pk, agent_pk)

    @staticmethod
    def affinity_variable():
        return 'a'

    def solve(self, verbose=False, max_compute_time=None, schedule_run=None):
        """ Return a schedule (if such one exists) as a list of (agent_pk, task_pk)
        :param verbose: print the result to stdout
        :param max_compute_time: max compute time
        :param schedule_run: a ScheduleRun object to update with its process id (must have its "id" attribute)
        :return:
        :rtype:
        """
        if max_compute_time is not None and max_compute_time <= 0:
            max_compute_time = None
        with tempfile.NamedTemporaryFile() as fd:
            for constraint in self.constraints():
                if verbose:
                    print(constraint)
                fd.write(('%s;\n' % constraint).encode())
            fd.flush()
            cmd = [settings.LP_SOLVE_PATH, '-lp', fd.name]
            if max_compute_time:
                cmd += ['-timeout', str(max_compute_time)]
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            if schedule_run:
                ScheduleRun.objects.filter(pk=schedule_run.pk).update(process_id=p.pid)
            std_out, std_err = p.communicate(timeout=max_compute_time)
            if schedule_run:
                ScheduleRun.objects.filter(pk=schedule_run.pk).update(process_id=None)
        if verbose:
            print(std_out.decode())
            print(std_err.decode())
        result_list = []
        std_out = std_out.decode()
        value_re = re.compile(self.result_line_re)
        for line in std_out.splitlines():
            matcher = value_re.match(line)
            if matcher:
                result_list.append((int(matcher.group(1)), int(matcher.group(2))))
        return result_list

    @staticmethod
    def result_by_agent(result_list):
        result_dict = {}
        for agent_pk, task_pk in result_list:
            result_dict.setdefault(agent_pk, set()).add(task_pk)
        return result_dict

    @staticmethod
    def result_by_task(result_list):
        result_dict = {}
        for agent_pk, task_pk in result_list:
            result_dict.setdefault(task_pk, set()).add(agent_pk)
        return result_dict



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-09-04 05:49
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('autoplanner', '0003_auto_20170903_2337'),
    ]

    operations = [
        migrations.AlterField(
            model_name='organization',
            name='current_schedule',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='current_organizations', to='autoplanner.ScheduleRun'),
        ),
    ]



import celery
import json
import os
import signal
import subprocess

import celery
from django.utils import timezone
from django.utils.formats import date_format, time_format
from django.utils.translation import ugettext_lazy as _
from djangofloor.celery import app
from djangofloor.signals.bootstrap3 import notify, SUCCESS, DANGER, INFO
from djangofloor.signals.html import render_to_client, content, after, replace_with
from djangofloor.wsgi.window_info import WindowInfo, render_to_string

from autoplanner.models import Organization, Task, ScheduleRun, Agent
from autoplanner.schedule import Scheduler

# used to avoid strange import bug with Python 3.2/3.3
# noinspection PyStatementEffect
celery.__file__
from celery import shared_task

__author__ = 'Matthieu Gallet'


class SetJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, set):
            return list(o)
        return super().default(o)


@shared_task(serializer='json', bind=True)
def compute_schedule(self, organization_id, window_info_data=None):
    window_info = None
    if window_info_data:
        window_info = WindowInfo.from_dict(window_info_data)
    celery_id = self.request.id
    start = timezone.localtime(timezone.now())
    msg = _('Computation started at %(d)s, %(t)s') % {'d': date_format(start, use_l10n=True),
                                                      't': time_format(start, use_l10n=True)}
    count = Organization.objects.filter(pk=organization_id, celery_task_id=None) \
        .update(celery_task_id=celery_id, celery_start=start, message=msg)
    if count == 0:
        return
    organization = Organization.objects.filter(pk=organization_id, celery_task_id=celery_id).first()
    if not organization:
        return
    if window_info:
        notify(window_info, content=str(msg), to=[organization], level=SUCCESS)
        organization.celery_start = start
        organization.celery_task_id = celery_id
        render_to_client(window_info, 'autoplanner/include/schedule_status.html', {'organization': organization},
                         '#schedule_status', to=[organization])
    schedule_run = ScheduleRun(organization=organization, celery_task_id=celery_id, celery_start=start)
    schedule_run.save()
    if window_info:
        content_str = render_to_string('autoplanner/include/schedule.html',
                                       context={'obj': schedule_run, 'organization': organization})
        after(window_info, '#schedules-header', content_str, to=[organization])
    scheduler = Scheduler(organization)
    level = SUCCESS
    try:
        result_list = scheduler.solve(verbose=False, max_compute_time=organization.max_compute_time,
                                      schedule_run=schedule_run)
        result_dict = scheduler.result_by_agent(result_list)
        end = timezone.localtime(timezone.now())
        selected = bool(result_dict)
        if selected:
            apply_schedule(organization_id, result_dict)
            msg = _('Computation finished at %(d)s, %(t)s') % {'d': date_format(end, use_l10n=True),
                                                               't': time_format(end, use_l10n=True)}
            balancing = scheduler.compute_balancing(result_list)
            if balancing:
                balances = _(', ').join(['%s: %s' % (x[0], x[2]) for x in balancing.values()])
                schedule_msg = _('Balancing: %(b)s') % {'b': balances}
            else:
                schedule_msg = _('No balance required.')
            ScheduleRun.objects.filter(organization__id=organization_id).update(is_selected=False)
        else:
            schedule_msg = _('Unable to find a solution')
            msg = _('Unable to find a solution, maybe you should remove some constraints or relax the balancing values.'
                    ' Computation finished at %(d)s, %(t)s') % {'d': date_format(end, use_l10n=True),
                                                                't': time_format(end, use_l10n=True)}
            level = DANGER
        serialized_result_dict = json.dumps(result_dict, cls=SetJSONEncoder)
        Organization.objects.filter(pk=organization_id).update(current_schedule_id=schedule_run.pk)
        ScheduleRun.objects.filter(pk=schedule_run.pk).update(celery_end=end, process_id=None, status=bool(result_dict),
                                                              result_dict=serialized_result_dict, is_selected=selected,
                                                              message=schedule_msg)
    except subprocess.TimeoutExpired:
        end = timezone.localtime(timezone.now())
        msg = _('%(d)s, %(t)s: Unable to find a schedule in the allowed time.') % \
            {'d': date_format(end, use_l10n=True), 't': time_format(end, use_l10n=True)}
        schedule_msg = _('Max computation time reached.')
        ScheduleRun.objects.filter(pk=schedule_run.pk).update(celery_end=end, process_id=None, status=False,
                                                              message=schedule_msg)
        level = DANGER
    except Exception as e:
        end = timezone.localtime(timezone.now())
        msg = _('%(d)s, %(t)s: Unable to compute a schedule %(msg)s') % \
            {'msg': e, 'd': date_format(end, use_l10n=True), 't': time_format(end, use_l10n=True)}
        schedule_msg = _('Unexpected error happened: %(e)s') % {'e': e}
        ScheduleRun.objects.filter(pk=schedule_run.pk).update(celery_end=end, process_id=None, status=False,
                                                              message=schedule_msg)
        level = DANGER
    Organization.objects.filter(pk=organization_id, celery_task_id=celery_id)\
        .update(celery_task_id=None, message=str(msg), celery_end=end)
    if window_info:
        notify(window_info, content=str(msg), to=[organization], level=level)
        organization.celery_end = end
        organization.celery_task_id = None
        render_to_client(window_info, 'autoplanner/include/schedule_status.html', {'organization': organization},
                         '#schedule_status', to=[organization])
        content_str = render_to_string('autoplanner/include/schedule.html',
                                       context={'obj': schedule_run, 'organization': organization})
        replace_with(window_info, '#schedule_%s' % schedule_run.id, content_str)


def apply_schedule(organization_id, result_dict):
    """Update all tasks to set the right agent"""
    available_agent_ids = {x[0] for x in Agent.objects.filter(organization__id=organization_id).values_list('id')}
    available_tasks_ids = {x[0] for x in Task.objects.filter(organization__id=organization_id).values_list('id')}
    all_task_pks = set()
    for agent_pk, task_pks in result_dict.items():
        agent_pk = int(agent_pk)
        task_pks = set(task_pks)
        if agent_pk not in available_agent_ids:
            raise ValueError(_('Invalid schedule: an agent has been removed since this computation'))
        elif not task_pks.issubset(available_tasks_ids):
            raise ValueError(_('Invalid schedule: a task has been removed since this computation'))
        all_task_pks |= task_pks
    if not available_tasks_ids.issubset(all_task_pks):
        raise ValueError(_('Invalid schedule: several tasks are not assigned to a resource'))
    for agent_pk, task_pks in result_dict.items():
        agent_pk = int(agent_pk)
        Task.objects.filter(pk__in=task_pks, fixed=False, organization__id=organization_id) \
            .update(agent_id=agent_pk)


@shared_task(serializer='json', bind=True)
def kill_schedule(self, celery_task_id, window_info_data=None):
    window_info = None
    if window_info_data:
        window_info = WindowInfo.from_dict(window_info_data)
    organization = Organization.objects.filter(celery_task_id=celery_task_id).first()
    if organization and window_info:
        notify(window_info, content=str(_('Cancel in progress…')), to=[organization], level=INFO)

    # noinspection PyUnusedLocal
    self = self
    app.control.revoke(celery_task_id)
    all_process_ids = []
    for values in ScheduleRun.objects.filter(celery_task_id=celery_task_id).values_list('process_id'):
        if values[0]:
            # noinspection PyBroadException
            try:
                os.kill(values[0], signal.SIGKILL)
                all_process_ids.append(values[0])
            except:
                pass
    end = timezone.localtime(timezone.now())
    ScheduleRun.objects.filter(celery_task_id=celery_task_id, process_id__in=all_process_ids).update(process_id=None,
                                                                                                     celery_end=end)
    msg = _('Computation killed at %(d)s, %(t)s') % {'d': date_format(end, use_l10n=True),
                                                     't': time_format(end, use_l10n=True)}
    Organization.objects.filter(celery_task_id=celery_task_id).update(celery_task_id=None, celery_end=end, message=msg)
    if organization and window_info:
        organization.celery_end = end
        organization.celery_task_id = None
        render_to_client(window_info, 'autoplanner/include/schedule_status.html', {'organization': organization},
                         '#schedule_status', to=[organization])
        notify(window_info, content=str(msg), to=[organization], level=SUCCESS)



from django import forms
from django.contrib.admin.widgets import AdminDateWidget, AdminSplitDateTime, AdminTimeWidget
from django.core.validators import RegexValidator
from django.utils.translation import ugettext_lazy as _

from autoplanner.models import Task, default_day_start, Category, MaxTaskAffectation, Agent
from autoplanner.utils import TimeDeltaField

__author__ = 'Matthieu Gallet'


class MultiplyTaskForm(forms.Form):
    source_task = forms.ModelChoiceField(queryset=Task.objects.all(),
                                         widget=forms.HiddenInput())
    until = forms.DateTimeField(label=_('Duplicate this task until'), initial=default_day_start,
                                widget=AdminSplitDateTime())
    every = forms.IntegerField(label=_('Days between successive task starts'), min_value=1, initial=7)


class OrganizationDescriptionForm(forms.Form):
    description = forms.CharField(required=False)


class OrganizationAccessTokenForm(forms.Form):
    access_token = forms.CharField(required=False, min_length=1, max_length=300,
                                   validators=[RegexValidator(r'^[a-zA-Z0-9]{1,300}$')])


class OrganizationMaxComputeTimeForm(forms.Form):
    max_compute_time = forms.IntegerField(required=False, min_value=0)


class CategoryNameForm(forms.Form):
    name = forms.CharField(label=_('Name'), max_length=500, min_length=1)


class CategoryBalancingModeForm(forms.Form):
    balancing_mode = forms.ChoiceField(label=_('Balancing mode'), required=False,
                                       choices=((None, _('No balancing')),
                                                (Category.BALANCE_TIME, _('Total task time')),
                                                (Category.BALANCE_NUMBER, _('Total task number'))))


class CategoryBalancingToleranceTimeForm(forms.Form):
    balancing_tolerance = TimeDeltaField(label=_('Tolerance while balancing tasks across resources'),
                                         required=False)


class CategoryBalancingToleranceNumberForm(forms.Form):
    balancing_tolerance = forms.FloatField(label=_('Tolerance while balancing tasks across resources'),
                                           required=False)


class CategoryAutoAffinityForm(forms.Form):
    auto_affinity = forms.FloatField(label=_('Affinity for allocating successive tasks of the same category '
                                             'to the same agent'))


class CategoryAddForm(forms.Form):
    name = forms.CharField(label=_('Name'), max_length=500, min_length=1)
    balancing_mode = forms.ChoiceField(label=_('Balancing mode'), required=False,
                                       choices=((None, _('No balancing')),
                                                (Category.BALANCE_TIME, _('Total task time')),
                                                (Category.BALANCE_NUMBER, _('Total task number'))))
    balancing_tolerance = forms.FloatField(label=_('Tolerance while balancing tasks across resources'),
                                           required=False)
    auto_affinity = forms.FloatField(label=_('Affinity for allocating successive tasks of the same category '
                                             'to the same agent'))


class AgentNameForm(forms.Form):
    name = forms.CharField(label=_('Name'), max_length=500, min_length=1)


class AgentStartDateForm(forms.Form):
    start_time_0 = forms.DateField(label=_('Arrival time'), required=False, widget=AdminDateWidget())


class AgentEndDateForm(forms.Form):
    end_time_0 = forms.DateField(label=_('Leaving time'), required=False, widget=AdminDateWidget())


class AgentStartTimeForm(forms.Form):
    start_time_1 = forms.TimeField(label=_('Arrival time'), required=False, widget=AdminTimeWidget())


class AgentEndTimeForm(forms.Form):
    end_time_1 = forms.TimeField(label=_('Leaving time'), required=False, widget=AdminTimeWidget())


class AgentAddForm(forms.Form):
    name = forms.CharField(label=_('Name'), max_length=500, min_length=1)


class AgentCategoryPreferencesAffinityForm(forms.Form):
    affinity = forms.FloatField(label=_('Affinity of the agent for the category.'), initial=0.0)


class AgentCategoryPreferencesBalancingOffsetForm(forms.Form):
    balancing_offset = forms.FloatField(label=_('Number of tasks already done'), initial=0.0)


class AgentCategoryPreferencesBalancingOffsetTimeForm(forms.Form):
    balancing_offset = TimeDeltaField(label=_('Time already passed on this category of tasks'), initial='0d')


class AgentCategoryPreferencesBalancingCountForm(forms.Form):
    balancing_count = forms.FloatField(label=_('If a task of this category performed by this agent counts twice, '
                                               'set this number to 2.0.'), initial=1.0, required=False,
                                       help_text=_('Blank if the agent cannot perform tasks of this category'))


class AgentCategoryPreferencesAddForm(forms.Form):
    category = forms.IntegerField()


class MaxTaskAffectationCategoryForm(forms.Form):
    category = forms.ModelChoiceField(queryset=Category.objects.all())


class MaxTaskAffectationModeForm(forms.Form):
    mode = forms.ChoiceField(choices=((MaxTaskAffectation.MINIMUM, _('At least this number of tasks')),
                                      (MaxTaskAffectation.MAXIMUM, _('At most this number of tasks'))),
                             initial=MaxTaskAffectation.MAXIMUM)


class MaxTaskAffectationRangeTimeSliceForm(forms.Form):
    range_time_slice = TimeDeltaField(label=_('Period length (days)'), initial='2d')


class MaxTaskAffectationTaskMaximumCountForm(forms.Form):
    task_maximum_count = forms.IntegerField(label=_('Number of tasks in this range'), initial=1)


class MaxTimeAffectationTaskMaximumTimeForm(forms.Form):
    task_maximum_time = TimeDeltaField(label=_('Number of tasks in this range'), initial='12:00')


class MaxTaskAffectationAddForm(forms.Form):
    category = forms.ModelChoiceField(queryset=Category.objects.all())
    mode = forms.ChoiceField(choices=((MaxTaskAffectation.MINIMUM, _('At least this number of tasks')),
                                      (MaxTaskAffectation.MAXIMUM, _('At most this number of tasks'))),
                             initial=MaxTaskAffectation.MAXIMUM)
    range_time_slice = TimeDeltaField(label=_('Period length (days)'), initial='2d')
    task_maximum_count = forms.IntegerField(label=_('Number of tasks in this range'), initial=1)


class MaxTimeAffectationAddForm(forms.Form):
    category = forms.ModelChoiceField(queryset=Category.objects.all())
    mode = forms.ChoiceField(choices=((MaxTaskAffectation.MINIMUM, _('At least this number of tasks')),
                                      (MaxTaskAffectation.MAXIMUM, _('At most this number of tasks'))),
                             initial=MaxTaskAffectation.MAXIMUM)
    range_time_slice = TimeDeltaField(label=_('Period length (days)'), initial='2d')
    task_maximum_time = TimeDeltaField(label=_('Number of tasks in this range'), initial='12:00')


class TaskNameForm(forms.Form):
    name = forms.CharField(label=_('Name'), max_length=500, min_length=1)


class TaskStartTimeForm(forms.Form):
    start_time_1 = forms.TimeField(label=_('Arrival time'), widget=AdminTimeWidget())


class TaskStartDateForm(forms.Form):
    start_time_0 = forms.DateField(label=_('Arrival time'), widget=AdminDateWidget())


class TaskEndTimeForm(forms.Form):
    end_time_1 = forms.TimeField(label=_('Arrival time'), widget=AdminTimeWidget())


class TaskEndDateForm(forms.Form):
    end_time_0 = forms.DateField(label=_('Arrival time'), widget=AdminDateWidget())


class TaskAgentForm(forms.Form):
    agent = forms.ModelChoiceField(queryset=Agent.objects.all(), required=False)


class TaskCategoriesForm(forms.Form):
    categories = forms.ModelMultipleChoiceField(queryset=Category.objects.all(), required=False)


class TaskAddForm(forms.Form):
    name = forms.CharField(label=_('Name'), max_length=500, min_length=1)
    start_time = forms.SplitDateTimeField(label=_('Starting time'), widget=AdminSplitDateTime())
    end_time = forms.SplitDateTimeField(label=_('End time'), widget=AdminSplitDateTime())
    agent = forms.ModelChoiceField(queryset=Agent.objects.all(), required=False)
    categories = forms.ModelMultipleChoiceField(queryset=Category.objects.all(), required=False)


class TaskMultiplyForm(forms.Form):
    until = forms.SplitDateTimeField(label=_('Duplicate this task until'), initial=default_day_start,
                                     widget=AdminSplitDateTime())
    every = forms.IntegerField(label=_('Days between successive task starts'), min_value=1, initial=7)


class TaskMultipleUpdateForm(forms.Form):
    agent = forms.ModelChoiceField(queryset=Agent.objects.all(), required=False)
    categories = forms.ModelMultipleChoiceField(queryset=Category.objects.all(), required=False)
    fix = forms.BooleanField(required=False)
    unfix = forms.BooleanField(required=False)
    tasks = forms.ModelMultipleChoiceField(queryset=Task.objects.all(), required=False)


class TaskMultipleRemoveForm(forms.Form):
    tasks = forms.ModelMultipleChoiceField(queryset=Task.objects.all(), required=False)


class TaskImportForm(forms.Form):
    csv_content = forms.CharField(widget=forms.Textarea())


class OrganizationAddForm(forms.Form):
    name = forms.CharField(label=_('Name'), max_length=500, min_length=1)



DF_JS = ['admin/js/core.js',
         'admin/js/calendar.js',
         'admin/js/admin/DateTimeShortcuts.js',
         'autoplanner/js/autoplanner.js']
DF_CSS = ['autoplanner/css/autoplanner.css']
DF_INDEX_VIEW = 'autoplanner.views.index'
DF_PROJECT_NAME = 'AutoPlanner'
DF_SITE_SEARCH_VIEW = None
# Make this unique, and don't share it with anybody.
LP_SOLVE_PATH = 'lp_solve'
REFRESH_DURATION = '1H'



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-09-03 21:37
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('autoplanner', '0002_auto_20170903_2258'),
    ]

    operations = [
        migrations.AlterField(
            model_name='schedulerun',
            name='celery_task_id',
            field=models.CharField(blank=True, db_index=True, default=None, max_length=60, null=True, verbose_name='Celery task id'),
        ),
    ]



from django.contrib import admin
from django.contrib import messages
from django.urls import reverse
from django.shortcuts import get_object_or_404
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _

from autoplanner.models import Task, AgentCategoryPreferences, AgentTaskExclusion, MaxTaskAffectation, Category, \
    Agent, Organization, MaxTimeTaskAffectation, ScheduleRun

__author__ = 'Matthieu Gallet'


class AgentInline(admin.TabularInline):
    model = Agent
    show_change_link = True
    classes = ['collapse', 'collapsed']


class CategoryInline(admin.TabularInline):
    model = Category
    classes = ['collapse', 'collapsed']

    def get_extra(self, request, obj=None, **kwargs):
        if request.GET.get('readonly'):
            return 0
        elif obj and obj.category_set.all().count() > 3:
            return 1
        return 3

    def get_readonly_fields(self, request, obj=None):
        if request.GET.get('readonly'):
            return ['name', 'balancing_mode', 'balancing_tolerance', 'auto_affinity']
        return []


class TaskInline(admin.TabularInline):
    model = Task
    classes = ['collapse', 'collapsed']

    @staticmethod
    def repeat(obj):
        if obj and obj.pk:
            return format_html('<a href="{}" class="button default">{}</a>',
                               reverse('multiply_task', kwargs={'task_pk': obj.pk}),
                               _('Repeat'))
        return ''

    repeat.allow_tags = True
    repeat.verbose_name = _('Repeat')
    readonly_fields = ('repeat',)
    fields = ('categories', 'name', 'start_time', 'end_time', 'agent', 'fixed', 'repeat')

    def get_readonly_fields(self, request, obj=None):
        if request.GET.get('readonly'):
            return ['categories', 'name', 'start_time', 'end_time', 'agent', 'fixed', 'repeat', ]
        return ['repeat', ]

    def get_extra(self, request, obj=None, **kwargs):
        if request.GET.get('readonly'):
            return 0
        return 3


class MaxTaskAffectationInline(admin.TabularInline):
    model = MaxTaskAffectation
    classes = ['collapse', 'collapsed']


class ScheduleRunInline(admin.TabularInline):
    model = ScheduleRun
    classes = ['collapse', 'collapsed']

    @staticmethod
    def apply(obj):
        if obj and obj.pk and obj.status:
            if obj.is_selected:
                return format_html('<span class="button default" style="background-color: #0B6138;">{}</span>',
                                   _('Current schedule'))
            return format_html('<a href="{}" class="button default">{}</a>',
                               reverse('apply_schedule_run', kwargs={'schedule_run_pk': obj.pk}),
                               _('Select this schedule'))
        return ''
    extra = 0
    apply.allow_tags = True
    apply.verbose_name = _('Reload this schedule')
    readonly_fields = ['status', 'message', 'celery_start', 'celery_end', 'apply']
    fields = ['status', 'message', 'celery_start', 'celery_end', 'apply']

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class MaxTimeTaskAffectationInline(admin.TabularInline):
    model = MaxTimeTaskAffectation
    classes = ['collapse', 'collapsed']


class OrganizationAdmin(admin.ModelAdmin):
    exclude = ['message', 'celery_task_id', 'celery_start', 'celery_end']
    list_display = ['name', 'message', 'access_token']

    def change_view(self, request, object_id, form_url='', extra_context=None):
        obj = get_object_or_404(Organization, pk=object_id)
        if obj.message and request.method == 'GET':
            messages.info(request, obj.message)
        return super().change_view(request, object_id, form_url=form_url, extra_context=extra_context)

    # noinspection PyMethodMayBeStatic
    def schedule_button(self, obj):
        if obj and obj.pk:
            if not obj.celery_task_id:
                return format_html('<a href="{}" class="button default">{}</a>',
                                   reverse('schedule_start', kwargs={'organization_pk': obj.pk}),
                                   _('Compute a new schedule!'))
            else:
                return format_html('<a href="{}" class="button default">{}</a>',
                                   reverse('cancel_schedule', kwargs={'organization_pk': obj.pk}),
                                   _('Cancel the currently running schedule'))
        return ''

    def get_queryset(self, request):
        return self.model.query(request)

    schedule_button.allow_tags = True
    schedule_button.short_description = _('Compute a complete schedule')
    readonly_fields = ('schedule_button',)
    fields = ['name', 'description', 'access_token', 'admins', 'schedule_button', 'max_compute_time', ]
    inlines = [ScheduleRunInline, AgentInline, CategoryInline, MaxTaskAffectationInline, MaxTimeTaskAffectationInline,
               TaskInline, ]


class AgentCategoryPreferencesInline(admin.StackedInline):
    model = AgentCategoryPreferences
    fields = ('category', 'agent', 'affinity', 'balancing_offset', 'balancing_count',)
    classes = ['collapse', 'collapsed']


class AgentTaskExclusionInline(admin.TabularInline):
    model = AgentTaskExclusion
    fields = ('agent', 'task',)


class AgentAdmin(admin.ModelAdmin):
    inlines = [AgentCategoryPreferencesInline, AgentTaskExclusionInline]
    list_display = ('name', 'organization', 'start_time', 'end_time',)
    list_filter = ('organization',)

    def get_queryset(self, request):
        return self.model.query(request)

    def get_fields(self, request, obj=None):
        fields = ['name', 'start_time', 'end_time']
        if obj is None:
            fields = ['organization'] + fields
        return fields


class TaskAdmin(admin.ModelAdmin):
    def get_queryset(self, request):
        return self.model.query(request)

    # noinspection PyMethodMayBeStatic
    def repeat_button(self, obj):
        if obj and obj.pk:
            return format_html('<a href="{}" class="button default">{}</a>',
                               reverse('multiply_task', kwargs={'task_pk': obj.pk}),
                               _('Repeat'))
        return ''

    def fix_agents(self, request, queryset):
        rows_updated = queryset.exclude(agent=None).update(fixed=True)
        if rows_updated > 1:
            self.message_user(request, _('%(t)s tasks have been updated') % {'t': rows_updated})
        else:
            self.message_user(request, _('%(t)s task has been updated') % {'t': rows_updated})

    fix_agents.short_description = _('Fix the agent processing the selected tasks ')

    def unfix_agents(self, request, queryset):
        rows_updated = queryset.exclude(agent=None).update(fixed=False)
        if rows_updated > 1:
            self.message_user(request, _('%(t)s tasks have been updated') % {'t': rows_updated})
        else:
            self.message_user(request, _('%(t)s task has been updated') % {'t': rows_updated})

    unfix_agents.short_description = _('Unfix the agent processing the selected tasks ')

    repeat_button.allow_tags = True
    readonly_fields = ('repeat_button',)
    repeat_button.short_description = _('Repeat')
    actions = ['fix_agents', 'unfix_agents']
    search_fields = ['name', ]
    fields = ['categories', 'name', 'start_time', 'end_time', 'agent', 'fixed', ]
    list_display = ['name', 'start_time', 'end_time', 'agent', 'fixed', 'repeat_button', ]
    list_editable = ['start_time', 'end_time', 'agent', 'fixed']
    list_filter = ['categories', 'agent', 'organization', 'fixed', 'start_time', 'end_time', ]


class AgentCategoryPreferencesAdmin(admin.ModelAdmin):
    list_display = ['category', 'agent', 'affinity', 'balancing_offset', 'balancing_count']
    list_filter = ['category', 'agent', 'organization', ]


admin.site.register(Organization, OrganizationAdmin)
admin.site.register(Agent, AgentAdmin)
admin.site.register(Task, TaskAdmin)
admin.site.register(AgentCategoryPreferences, AgentCategoryPreferencesAdmin)
# admin.site.unregister(Site)
admin.site.site_header = _('AutoPlanner')



from django.conf.urls import url

from autoplanner import views

__author__ = 'Matthieu Gallet'
app_name = 'autoplanner'
urlpatterns = [
    url(r'^organization/(?P<organization_pk>\d+)/$', views.organization_index, name='organization_index'),

    url(r'^org/(?P<organization_pk>\d+)\.html', views.organization, name='organization'),
    url(r'^org/start/(?P<organization_pk>\d+)\.html', views.schedule_task, name='schedule_start'),
    url(r'^org/cancel/(?P<organization_pk>\d+)\.html', views.cancel_schedule_task, name='cancel_schedule'),
    url(r'^org/ical/(?P<organization_pk>\d+)/all/(?P<title>.*)\.ics', views.generate_ics, name='ical'),
    url(r'^org/ical/(?P<organization_pk>\d+)/agent/(?P<agent_pk>\d+)/(?P<title>.*)\.ics',
        views.generate_ics, name='ical'),
    url(r'^org/ical/(?P<organization_pk>\d+)/cat/(?P<category_pk>\d+)/(?P<title>.*)\.ics',
        views.generate_ics, name='ical'),
    url(r'^tasks/multiply/(?P<task_pk>\d+)\.html', views.multiply_task, name='multiply_task'),
    url(r'^tasks/apply_schedule_run/(?P<schedule_run_pk>\d+)\.html', views.apply_schedule_run,
        name='apply_schedule_run'),
]



import codecs
import os.path
import re

from setuptools import setup, find_packages


version = None
for line in codecs.open(os.path.join('autoplanner', '__init__.py'), 'r', encoding='utf-8'):
    matcher = re.match(r"""^__version__\s*=\s*['"](.*)['"]\s*$""", line)
    version = version or matcher and matcher.group(1)

# get README content from README.md file
with codecs.open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf-8') as fd:
    long_description = fd.read()

entry_points = {'console_scripts': ['autoplanner-manage = djangofloor.scripts:django',
                                    'autoplanner-celery = djangofloor.scripts:celery',
                                    'autoplanner-web = djangofloor.scripts:aiohttp']}

setup(
    name='autoplanner',
    version=version,
    description='No description yet.',
    long_description=long_description,
    author='Matthieu Gallet',
    author_email='matthieu.gallet@19pouces.net',
    license='CeCILL-B',
    url='http://autoplanner.readthedocs.org/en/latest/',
    entry_points=entry_points,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    test_suite='autoplanner.tests',
    install_requires=['djangofloor', 'icalendar', 'markdown', ],
    setup_requires=[],
    classifiers=['Development Status :: 3 - Alpha', 'Operating System :: MacOS :: MacOS X',
                 'Operating System :: Microsoft :: Windows', 'Operating System :: POSIX :: BSD',
                 'Operating System :: POSIX :: Linux', 'Operating System :: Unix',
                 'License :: OSI Approved :: CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)',
                 'Programming Language :: Python :: 3.4', 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3 :: Only'],
)



from django import template
from django.forms import TimeInput, DateInput
from django.utils import formats
from django.utils.safestring import mark_safe
from markdown import markdown as mkdown
from autoplanner.utils import python_to_str

__author__ = 'Matthieu Gallet'
register = template.Library()


@register.filter
def my_simple_str(x):
    """Seems to be useless, but "no signature found for builtin type <class 'str'>" is raised otherwise"""
    return str(x)


@register.filter
def timedelta(x):
    return mark_safe(python_to_str(x))


@register.filter
def input_time(value):
    if value is None:
        return mark_safe('')
    return mark_safe(formats.localize_input(value, formats.get_format(TimeInput.format_key)[0]))


@register.filter
def int_format(value, fmt=''):
    if value is None:
        return mark_safe('')
    return mark_safe(('%%%sd' % fmt) % value)


@register.filter
def input_date(value):
    if value is None:
        return mark_safe('')
    return mark_safe(formats.localize_input(value, formats.get_format(DateInput.format_key)[0]))


@register.filter
def js_value(value):
    if value is None:
        return mark_safe('null')
    return mark_safe(repr(value))


@register.filter
def markdown(value):
    return mark_safe(mkdown(value))



import datetime
import json
from django.conf import settings
from django.template.response import TemplateResponse
from django.utils.formats import date_format
from django.utils.formats import time_format
from django.views.decorators.cache import never_cache
import re

from django.contrib import admin
from django.contrib.admin.options import IS_POPUP_VAR, get_content_type_for_model
from django.contrib.admin.options import TO_FIELD_VAR
from django.contrib.admin.templatetags.admin_urls import add_preserved_filters
from django.contrib import messages
from django.contrib.admin.utils import quote
from django.urls import reverse
from django.http.response import HttpResponseRedirect, HttpResponse
from django.shortcuts import render_to_response, get_object_or_404
from django.template import RequestContext
from django.utils.translation import ugettext_lazy as _
from djangofloor.celery import app
from djangofloor.tasks import set_websocket_topics
from icalendar import Calendar, Event
import markdown

from autoplanner.admin import OrganizationAdmin
from autoplanner.forms import MultiplyTaskForm, OrganizationAddForm
from autoplanner.models import Organization, Task, Category, Agent, API_KEY_VARIABLE, ScheduleRun
from autoplanner.schedule import Scheduler
from autoplanner.tasks import compute_schedule, kill_schedule, apply_schedule

__author__ = 'Matthieu Gallet'


def get_template_values(request, organization_pk):
    org = get_object_or_404(Organization.query(request, readonly=True), pk=organization_pk)
    assert isinstance(org, Organization)
    # noinspection PyProtectedMember
    model_admin = admin.site._registry[Organization]
    assert isinstance(model_admin, OrganizationAdmin)
    # noinspection PyProtectedMember
    opts = model_admin.model._meta
    app_label = opts.app_label
    preserved_filters = model_admin.get_preserved_filters(request)
    form_url = add_preserved_filters({'preserved_filters': preserved_filters, 'opts': opts}, '?')
    view_on_site_url = model_admin.get_view_on_site_url(org)
    template_values = {
        'is_popup': 0,
        'add': False,
        'change': True,
        'has_change_permission': model_admin.has_change_permission(request, org),
        'has_delete_permission': False,
        'has_add_permission': False,
        'has_file_field': False,
        'has_absolute_url': view_on_site_url is not None,
        'absolute_url': view_on_site_url,
        'form_url': form_url,
        'opts': opts,
        'content_type_id': get_content_type_for_model(Task).pk,
        'save_as': True,
        'save_on_top': False,
        'media': model_admin.media,
        'to_field_var': TO_FIELD_VAR,
        'is_popup_var': IS_POPUP_VAR,
        'app_label': app_label,
        'show_save': False,
    }
    return template_values


@never_cache
def organization(request, organization_pk):
    template_values = get_template_values(request, organization_pk)
    obj = get_object_or_404(Organization.query(request, readonly=True), pk=organization_pk)
    scheduler = Scheduler(obj)
    messages.info(request, obj.message)
    statistics = {x.pk: {category.pk: [0, datetime.timedelta(0), None] for category in scheduler.categories}
                  for x in scheduler.agents}
    # [total_number, total_time, balanced_value]
    for category in scheduler.categories:
        if category.balancing_mode == Category.BALANCE_NUMBER:
            for agent in scheduler.agents:
                offset, count, __ = scheduler.preferences_by_agent_by_category[category.pk].get(agent.pk, (0., 1., 0.))
                if count is None:
                    statistics[agent.pk][category.pk][2] = None
                else:
                    statistics[agent.pk][category.pk][2] = offset * count
        elif category.balancing_mode == Category.BALANCE_TIME:
            for agent in scheduler.agents:
                offset, count, __ = scheduler.preferences_by_agent_by_category[category.pk].get(agent.pk, (0., 1., 0.))
                if count is None:
                    statistics[agent.pk][category.pk][2] = None
                else:
                    statistics[agent.pk][category.pk][2] = datetime.timedelta(seconds=offset * count)
    for task in scheduler.tasks:
        if task.agent_id is None:
            continue
        for category_pk in scheduler.categories_by_task[task.pk]:
            duration = task.duration
            statistics[task.agent_id][category_pk][0] += 1
            statistics[task.agent_id][category_pk][1] += duration
            if scheduler.categories_by_pk[category_pk].balancing_mode is None:
                continue
            __, count, __ = scheduler.preferences_by_agent_by_category[category_pk].get(task.agent_id, (0., 1., 0.))
            if scheduler.categories_by_pk[category_pk].balancing_mode == Category.BALANCE_NUMBER:
                if count is None:
                    statistics[task.agent_id][category_pk][2] = None
                else:
                    statistics[task.agent_id][category_pk][2] += count

            elif scheduler.categories_by_pk[category_pk].balancing_mode == Category.BALANCE_TIME:
                value = datetime.timedelta(seconds=duration.total_seconds() * count)
                statistics[task.agent_id][category_pk][2] += value

    categories = [category for category in scheduler.categories]
    categories.sort(key=lambda x: x.name)
    agents = [agent for agent in scheduler.agents]
    agents.sort(key=lambda x: x.name)
    sorted_statistics = [[x] + [statistics[x.pk][y.pk] for y in categories] for x in agents]
    template_values.update({
        'obj': obj,
        'statistics': sorted_statistics,
        'agents': agents,
        'categories': categories,
        'description': markdown.markdown(obj.description),
        'api_key_variable': API_KEY_VARIABLE,
    })
    return render_to_response('autoplanner/organization.html', template_values, RequestContext(request))


@never_cache
def multiply_task(request, task_pk):
    obj = get_object_or_404(Task.query(request), pk=task_pk)
    # noinspection PyProtectedMember
    model_admin = admin.site._registry[Organization]
    assert isinstance(model_admin, OrganizationAdmin)
    # noinspection PyProtectedMember
    opts = model_admin.model._meta
    if request.method == 'POST':
        form = MultiplyTaskForm(request.POST)
        if form.is_valid():
            until_src = form.cleaned_data['until']
            every_src = form.cleaned_data['every']
            assert isinstance(until_src, datetime.datetime)
            limit = datetime.datetime(year=until_src.year, month=until_src.month, day=until_src.day,
                                      hour=23, minute=59, second=59, tzinfo=obj.start_time.tzinfo)
            to_create = []
            increment = datetime.timedelta(days=every_src)
            start_time = obj.start_time + increment
            end_time = obj.end_time + increment
            matcher = re.match(r'^(.*)\s+\((\d+)\)', obj.name)
            if matcher:
                new_name = matcher.group(1)
                name_index = int(matcher.group(2)) + 1
            else:
                new_name = obj.name
                name_index = 2
            current_category_pks = [x.pk for x in obj.categories.all()]
            while start_time < limit:
                new_task = Task(organization_id=obj.organization_id, name='%s (%d)' % (new_name, name_index),
                                start_time=start_time, end_time=end_time)
                to_create.append(new_task)
                start_time += increment
                end_time += increment
                name_index += 1
            if to_create:
                if current_category_pks:
                    cls = Task.categories.through
                    all_categories_to_create = []
                    for new_task in to_create:
                        new_task.save()
                        all_categories_to_create += [cls(task_id=new_task.pk, category_id=category_pk)
                                                     for category_pk in current_category_pks]
                    cls.objects.bulk_create(all_categories_to_create)
                else:
                    Task.objects.bulk_create(to_create)
            if len(to_create) > 1:
                messages.info(request, _('%(count)d tasks have been created.') % {'count': len(to_create)})
            elif to_create:
                messages.info(request, _('A task has been created.'))
            new_url = reverse('admin:%s_%s_change' % (opts.app_label, opts.model_name),
                              args=(quote(obj.organization_id), ), current_app=model_admin.admin_site.name)
            return HttpResponseRedirect(new_url)
    else:
        form = MultiplyTaskForm(initial={'source_task': obj})
    template_values = get_template_values(request, organization_pk=obj.organization_id)
    template_values['obj'] = obj
    template_values['form'] = form
    return render_to_response('autoplanner/multiply_task.html', template_values, RequestContext(request))


@never_cache
def schedule_task(request, organization_pk):
    obj = get_object_or_404(Organization.query(request), pk=organization_pk)
    # noinspection PyProtectedMember
    model_admin = admin.site._registry[Organization]
    assert isinstance(model_admin, OrganizationAdmin)
    # noinspection PyProtectedMember
    opts = model_admin.model._meta
    count = Organization.objects.filter(pk=organization_pk, celery_task_id=None)
    if count == 0:
        messages.error(request, _('%(obj)s is already busy.') % {'obj': obj})
    else:
        compute_schedule.delay(obj.pk)
        messages.info(request, _('Computation launched.') % {'obj': obj})
    new_url = reverse('admin:%s_%s_change' % (opts.app_label, opts.model_name),
                      args=(quote(obj.pk), ), current_app=model_admin.admin_site.name)
    return HttpResponseRedirect(new_url)


@never_cache
def apply_schedule_run(request, schedule_run_pk):
    obj = get_object_or_404(ScheduleRun, pk=schedule_run_pk)
    organization_id = obj.organization_id
    get_object_or_404(Organization.query(request), pk=organization_id)  # only used to check rights
    # required to get the URL of the "administration change page"
    # noinspection PyProtectedMember
    model_admin = admin.site._registry[Organization]
    assert isinstance(model_admin, OrganizationAdmin)
    # noinspection PyProtectedMember
    opts = model_admin.model._meta

    result_dict = json.loads(obj.result_dict)
    end = obj.celery_end
    d = '%(d)s, %(t)s' % {'d': date_format(end, use_l10n=True), 't': time_format(end, use_l10n=True)}
    if not result_dict:
        messages.error(request, _('Unable to apply the invalid schedule "%(d)s".') % {'d': d})
    else:
        try:
            apply_schedule(organization_id, result_dict)
            Organization.objects.filter(pk=organization_id).update(current_schedule=obj.pk)
            ScheduleRun.objects.filter(organization__id=organization_id).exclude(pk=schedule_run_pk)\
                .update(is_selected=False)
            ScheduleRun.objects.filter(pk=schedule_run_pk).update(is_selected=True)
            messages.success(request, _('Schedule "%(d)s" has been applied.') % {'d': d})
        except ValueError as e:
            messages.error(request, _('Unable to apply the invalid schedule "%(d)s": %(e)s.') % {'d': d, 'e': e})

    new_url = reverse('admin:%s_%s_change' % (opts.app_label, opts.model_name),
                      args=(quote(organization_id),), current_app=model_admin.admin_site.name)
    return HttpResponseRedirect(new_url)


@never_cache
def cancel_schedule_task(request, organization_pk):
    obj = get_object_or_404(Organization.query(request), pk=organization_pk)
    # noinspection PyProtectedMember
    model_admin = admin.site._registry[Organization]
    assert isinstance(model_admin, OrganizationAdmin)
    # noinspection PyProtectedMember
    opts = model_admin.model._meta
    if obj.celery_task_id is None:
        messages.error(request, _('No computation is running on %(obj)s.') % {'obj': obj})
    else:
        kill_schedule.delay(obj.celery_task_id)
        messages.info(request, _('Computation will be interrupted.') % {'obj': obj})
    new_url = reverse('admin:%s_%s_change' % (opts.app_label, opts.model_name),
                      args=(quote(obj.pk), ), current_app=model_admin.admin_site.name)
    return HttpResponseRedirect(new_url)


@never_cache
def generate_ics(request, organization_pk, agent_pk=None, category_pk=None, title=''):
    obj = get_object_or_404(Organization.query(request, readonly=True), pk=organization_pk)
    cal = Calendar()
    cal.add('prodid', '-//AutoPlanner//19pouces.net//')
    cal.add('version', '2.0')
    cal.add('X-PUBLISHED-TTL', 'PT' + settings.REFRESH_DURATION)
    cal.add('X-WR-TIMEZONE', settings.TIME_ZONE)
    cal.add('X-WR-CALNAME', title or obj.name)
    cal.add('X-WR-CALDESC', obj.description)
    query = Task.objects.filter(organization=obj)
    if agent_pk:
        query = query.filter(agent__id=agent_pk)
    if category_pk:
        query = query.filter(categories__id=category_pk)
    agents = {x.pk: x.name for x in Agent.objects.filter(organization=obj)}
    for task in query:
        event = Event()
        if task.agent_id:
            summary = '%s (%s)' % (task.name, agents[task.agent_id])
        else:
            summary = task.name
        event.add('summary', summary)
        event.add('dtstart', task.start_time)
        event.add('dtend', task.end_time)
        event['uid'] = task.start_time.strftime('%Y%m%dT%H%M%S-') + str(task.pk)
        cal.add_component(event)
    return HttpResponse(cal.to_ical(), content_type='text/calendar')


def index(request):
    query = Organization.query(request).order_by('name')
    set_websocket_topics(request)
    if request.method == 'POST' and request.user.is_authenticated:
        form = OrganizationAddForm(request.POST)
        if form.is_valid():
            obj = Organization(name=form.cleaned_data['name'], max_compute_time=1800)
            obj.save()
            obj.admins = [request.user]
            messages.success(request, _('Organization successfully created.'))
            return HttpResponseRedirect(reverse('index'))
    else:
        form = OrganizationAddForm()
    template_values = {'organizations': query, 'form': form}
    return TemplateResponse(request, 'autoplanner/organizations.html', context=template_values)


def organization_index(request, organization_pk):
    obj = get_object_or_404(Organization.query(request), pk=organization_pk)
    set_websocket_topics(request, obj)
    template_values = {'organization': obj}
    return TemplateResponse(request, 'autoplanner/organization_index.html', context=template_values)
